using System;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using System.Runtime.Remoting;
using System.Runtime.Remoting.Messaging;
using System.Runtime.Remoting.Proxies;
using System.Reflection;
using System.Collections;
using System.IO;
using System.Runtime.InteropServices;

namespace DeviceHook
{

	/// <summary>
	/// Win32 API PerformanceCounter Wrapper
	/// </summary>
	/// <remarks>http://support.microsoft.com/default.aspx?scid=kb;ja;306979</remarks>
	public sealed class HighResolutionTimer
	{
		[System.Security.SuppressUnmanagedCodeSecurity]
		[DllImport("kernel32.dll")]
		private extern static short QueryPerformanceCounter(out long x);
		[System.Security.SuppressUnmanagedCodeSecurity]
		[DllImport("kernel32.dll")]
		private extern static short QueryPerformanceFrequency(out long x);

		private static readonly long PerformanceFrequency;

		/// <summary>
		/// ̃VXeňEx100imbPʂŕԂ
		/// </summary>
		public long MaximumResolution
		{
			get
			{
				return 10000000 / PerformanceFrequency;
			}
		}

		/// <summary>
		/// staticRXgN^
		/// </summary>
		static HighResolutionTimer()
		{
			short ret = QueryPerformanceFrequency(out PerformanceFrequency);
			if( ret == 0 )
			{
				throw new NotSupportedException("QueryPerformanceFrequency is not supported on this system.");
			}
		}

		/// <summary>
		/// VXeŇoߎԂԂ
		/// </summary>
		public static TimeSpan UpTime
		{
			get
			{
				long count;
				QueryPerformanceCounter(out count);
				long quotient = count / PerformanceFrequency;
				long rem = count - PerformanceFrequency * quotient;
				long ticks = quotient * 10000000 + (rem * 10000000) / PerformanceFrequency;
				return TimeSpan.FromTicks( ticks );
			}
		}
	}

	/// <summary>
	/// Microsoft.DirectX.Direct3D.Device ɑ΂vLV
	/// </summary>
	public class DeviceProxy : RealProxy
	{

		/// <summary>
		/// Draw`Primitives \bhɊւ
		/// </summary>
		private struct DrawEventInfo
		{
			public string MethodName;
			public long   StartTime;
			public long   EndTime;
			public int    PrimitiveCount;
			public override string ToString()
			{
				return String.Format( "{0},{1},{2},{3}", MethodName, StartTime, EndTime, PrimitiveCount );
			}

		}

		/// <summary>
		/// Present \bhɊւ
		/// </summary>
		private struct PresentEventInfo
		{
			public string MethodName;
			public long   StartTime;
			public long   EndTime;
			public override string ToString()
			{
				return String.Format( "{0},{1},{2}", MethodName, StartTime, EndTime );
			}		
		}
			
		static public Device CreateDevice(
			int adapter,
			DeviceType type,
			Control window,
			CreateFlags flags,
			PresentParameters param )
		{
			Device target = new Device( adapter, type, window, flags, param );

			DeviceProxy proxy = new DeviceProxy( target, typeof(Device) );
			return (Device) proxy.GetTransparentProxy();			
		}
		static public Device CreateDevice(
			int adapter,
			DeviceType type,
			IntPtr handle,
			CreateFlags flags,
			PresentParameters param )
		{
			Device target = new Device( adapter, type, handle, flags, param );
			DeviceProxy proxy = new DeviceProxy( target, typeof(Device) );
			return (Device) proxy.GetTransparentProxy();			
		}

		private readonly MarshalByRefObject target;
		private readonly ArrayList          events = new ArrayList();
		private readonly Hashtable          primitiveCountArg = new Hashtable();
		public DeviceProxy( MarshalByRefObject target, Type type) : base(type)
		{
			this.target = target;

			primitiveCountArg[ "DrawPrimitives" ] = 2;
			primitiveCountArg[ "DrawUserPrimitives" ] = 1;
			primitiveCountArg[ "DrawIndexedPrimitives" ] = 5;
			primitiveCountArg[ "DrawIndexedUserPrimitives" ] = 3;
		}
		public override IMessage Invoke(IMessage msg)
		{
			// Call bZ[W
			IMethodCallMessage call = (IMethodCallMessage) msg;

			switch( call.MethodName )
			{

				case "DrawPrimitives":
				case "DrawUserPrimitives":
				case "DrawIndexedPrimitives":
				case "DrawIndexedUserPrimitives":
				{
					DrawEventInfo eventInfo = new DrawEventInfo();
					eventInfo.MethodName = call.MethodName;
					eventInfo.StartTime  = HighResolutionTimer.UpTime.Ticks;
					IMethodReturnMessage ret =  ExecuteMessage( call );
					eventInfo.EndTime  = HighResolutionTimer.UpTime.Ticks;
					eventInfo.PrimitiveCount = (int) call.InArgs[ (int) primitiveCountArg[call.MethodName] ];
					events.Add( eventInfo );
					return ret;
				}
				case "Present":
				{
					PresentEventInfo eventInfo = new PresentEventInfo();
					eventInfo.MethodName = call.MethodName;
					eventInfo.StartTime  = HighResolutionTimer.UpTime.Ticks;
					IMethodReturnMessage ret =  ExecuteMessage( call );
					eventInfo.EndTime  = HighResolutionTimer.UpTime.Ticks;
					events.Add( eventInfo );
					return ret;
				}
				case "Dispose":
				{
					// ŃCxgf[^_v
					StreamWriter sw = new StreamWriter( "DeviceEvent.txt" );

					foreach( object info in events )
					{
						sw.WriteLine( info.ToString() );
					}
					sw.Close();

					return ExecuteMessage( call );
				}
				default:
					return ExecuteMessage( call );
			}

		}

		public unsafe IMethodReturnMessage ExecuteMessage( IMethodCallMessage call )
		{
			if( call.MethodName == "FieldGetter" && call.InArgCount > 2 && ((string)call.InArgs[1]) == "m_lpUM" )
			{
				// ̂Ƃ RemotingServices.ExecuteMessage ɔCƎŝŏꓖIɑΉ
				Type type = typeof(Device);
				FieldInfo info = type.GetField( "m_lpUM", BindingFlags.NonPublic | BindingFlags.Instance);
				object retval = info.GetValue( target );
				unsafe
				{
					void* rawptr = System.Reflection.Pointer.Unbox( retval );
					IntPtr ptr = new IntPtr( rawptr );
					return new ReturnMessage( typeof(void), new object[]{null,null,ptr}, 3, call.LogicalCallContext, call );
				}
			}
			else
			{
				return RemotingServices.ExecuteMessage( target, call );
			}
		}
	}
}
